/**
 * Copyright (C) 2010-2011 J.W.Marsden
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/
package jsonij.json;

import java.io.IOException;
import java.io.Reader;
import java.lang.reflect.Array;
import java.util.ArrayList;
import jsonij.parser.ParserException;

/**
 *
 * @author openecho
 */
public class JPath<C extends JPath.Component> extends ArrayList<C> {

    boolean recordEvaluateTime;
    long lastEvaluateTime;

    public JPath() {
        recordEvaluateTime = false;
        lastEvaluateTime = -1;
    }

    public Value evaluate(JSON json) throws JPathRuntimeException {
        Value value = json.getRoot();
        return evaluate(value);
    }

    public Value[] evaluateAll(JSON json) throws JPathRuntimeException {
        Value value = json.getRoot();
        return evaluateAll(value);
    }

    public Value evaluate(Value value) throws JPathRuntimeException {
        Value[] returnValues = evaluateAll(value);
        if(Array.getLength(returnValues) > 0) {
            return returnValues[0];
        } else {
            return null;
        }
    }

    public Value[] evaluateAll(Value value) throws JPathRuntimeException {
        long startTime = 0, finishTime = -1;
        if (recordEvaluateTime) {
            startTime = System.nanoTime();
        }

        Value valueStore;
        ArrayList<Value> currentValues = new ArrayList<Value>();
        ArrayList<Value> nextValues = null;
        currentValues.add(value);

        JPath.Key key;
        int count = 0;

        for (int i=0;i<size();i++) {
            JPath.Component c = get(i);
            nextValues = new ArrayList<Value>();
            if (c.getClass() == Key.class) {
                key = (JPath.Key) c;
                for(Value currentValue: currentValues) {

                    if((valueStore = currentValue.get(key.getValue())) != null) {
                        nextValues.add(valueStore);
                    }
                }
            } else if (c.getClass() == SimpleIndex.class) {
                SimpleIndex predicate = (SimpleIndex) c;
                if (predicate.isLast()) {
                    int indexPredicate = -1;
                    for(Value currentValue: currentValues) {
                        if ((indexPredicate = currentValue.size() - 1) >= 0) {
                            nextValues.add(currentValue.get(indexPredicate));
                        }
                    }
                } else {
                    for(Value currentValue: currentValues) {
                        if(predicate.getIndex() < currentValue.size()) {
                            nextValues.add(currentValue.get(predicate.getIndex()));
                        }
                    }
                }
            } else if (c.getClass() == LastIndex.class) {
                LastIndex predicate = (LastIndex) c;
                if (!predicate.hasOffset()) {
                    int indexPredicate = -1;
                    for(Value currentValue: currentValues) {
                        if ((indexPredicate = currentValue.size() - 1) >= 0) {
                            nextValues.add(currentValue.get(indexPredicate));
                        }
                    }
                } else {
                    int indexPredicate = -1;
                    for(Value currentValue: currentValues) {
                        if ((indexPredicate = currentValue.size() - predicate.getOffset()) >= 0) {
                            nextValues.add(currentValue.get(indexPredicate));
                        }
                    }
                }
            } else if (c.getClass() == ExpressionPredicate.class) {
                ExpressionPredicate predicate = (ExpressionPredicate) c;
                ArrayList<ExpressionPredicateCondition> conditions = predicate.conditions();
                for(Value currentValue: currentValues) {
                    if(currentValue.type() == Value.TYPE.ARRAY) {
                        for(int j=0;j<currentValue.size();j++) {
                            Value checkElement = currentValue.get(j);
                            if(checkElement.type() == Value.TYPE.OBJECT) {
                                boolean expressionValid = true;
                                for(ExpressionPredicateCondition condition: conditions) {
                                    Value checkValue = checkElement.get(condition.getAttribute());
                                    if(checkValue == null) {
                                        expressionValid = false;
                                        break;
                                    }
                                    ExpressionPredicateOperator expressionConditionOperator = condition.getOperator();
                                    if(expressionConditionOperator.equals(JPath.ExpressionPredicateOperator.NOT_NULL)) {
                                        if(checkValue.isNull()) {
                                            expressionValid = false;
                                            break;
                                        }
                                    } else if(expressionConditionOperator.equals(JPath.ExpressionPredicateOperator.EQUAL)) {
                                        if(!checkValue.equals(condition.value)) {
                                            expressionValid = false;
                                            break;
                                        }
                                    } else if(expressionConditionOperator.equals(JPath.ExpressionPredicateOperator.LESS)) {

                                    } else if(expressionConditionOperator.equals(JPath.ExpressionPredicateOperator.GREATER)) {

                                    } else if(expressionConditionOperator.equals(JPath.ExpressionPredicateOperator.LESS_EQUAL)) {

                                    } else if(expressionConditionOperator.equals(JPath.ExpressionPredicateOperator.GREATER_EQUAL)) {

                                    }

                                }

                                if(expressionValid) {
                                    nextValues.add(checkElement);
                                }
                            }
                        }
                    }
                }
            }
            currentValues = nextValues;
            count++;
        }

        if (recordEvaluateTime) {
            finishTime = System.nanoTime();
            lastEvaluateTime = finishTime - startTime;
        }

        int size = currentValues.size();
        Value[] returnValues = new Value[size];
        for(int i=0;i<size;i++) {
            returnValues[i]=currentValues.get(i);
        }
        return returnValues;
    }

    public boolean isRecordEvaluateTime() {
        return recordEvaluateTime;
    }

    public void setRecordEvaluateTime(boolean recordEvaluateTime) {
        this.recordEvaluateTime = recordEvaluateTime;
    }

    public long getLastEvaluateTime() {
        return lastEvaluateTime / 1000L;
    }

    public static JPath<Component> parse(String path) throws IOException, ParserException {
        JPathParser parser = new JPathParser();
        JPath<Component> jPath = parser.parse(path);
        return jPath;
    }

    public static Value evaluate(java.lang.String document, java.lang.String jPath) throws ParserException, IOException {
        JSON jsonDocument = JSON.parse(document);
        JPath jPathInstance = JPath.parse(jPath);
        return jPathInstance.evaluate(jsonDocument);
    }

    public static Value evaluate(Reader documentReader, java.lang.String jPath) throws ParserException, IOException {
        JSON jsonDocument = JSON.parse(documentReader);
        JPath jPathInstance = JPath.parse(jPath);
        return jPathInstance.evaluate(jsonDocument);
    }

    public static abstract class Component {
    }

    public static class Key extends Component {

        String value;

        public Key(String value) {
            this.value = value;
        }

        public String getValue() {
            return value;
        }

        public void setValue(String value) {
            this.value = value;
        }

        @Override
        public String toString() {
            return "Key{" + "value=" + value + '}';
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final Key other = (Key) obj;
            return this.value.equals(other.value);
        }

        @Override
        public int hashCode() {
            int hash = 5;
            hash = 53 * hash + (this.value != null ? this.value.hashCode() : 0);
            return hash;
        }
    }

    public static abstract class Predicate extends Component {
    }

    /**
     * SimpleIndex Predicate Implementation. The most basic of the Predicates.
     *
     * @author openecho
     */
    public static class SimpleIndex extends JPath.Predicate {

        public static final int LAST_INDEX;
        int index;

        static {
            LAST_INDEX = -1;
        }

        public SimpleIndex() {
            index = 0;
        }

        public SimpleIndex(int index) {
            this.index = index;
        }

        public boolean isLast() {
            return index <= LAST_INDEX;
        }

        public void setLast() {
            this.index = LAST_INDEX;
        }

        public int getIndex() {
            return index;
        }

        public void setIndex(int index) {
            this.index = index;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final SimpleIndex other = (SimpleIndex) obj;
            if (this.index != other.index) {
                return false;
            }
            return true;
        }

        @Override
        public int hashCode() {
            int hash = 5;
            hash = 53 * hash + this.index;
            return hash;
        }
    }

    /**
     * LastIndex Predicate Implementation.
     *
     * @author openecho
     */
    public static class LastIndex extends JPath.Predicate {

        int offset;
        public static final LastIndex LAST;

        static {
            LAST = new LastIndex();
        }

        public LastIndex() {
            this(0);
        }

        public LastIndex(int offset) {
            this.offset = Math.abs(offset);
        }

        public boolean hasOffset() {
            return offset != 0;
        }

        public int getOffset() {
            return offset;
        }

        public void setOffset(int offset) {
            this.offset = offset;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final LastIndex other = (LastIndex) obj;
            if (this.offset != other.offset) {
                return false;
            }
            return true;
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 71 * hash + this.offset;
            return hash;
        }
    }

    /**
     * Complex Predicate Implementation.
     *
     * @author openecho
     */
    public static class ExpressionPredicate extends JPath.Predicate {

        String expression;
        ArrayList<ExpressionPredicateCondition> conditions;

        static {
       
        }

        public ExpressionPredicate() {
            this.expression = null;
            this.conditions = new ArrayList<ExpressionPredicateCondition>();
        }

        public ArrayList<ExpressionPredicateCondition> getConditions() {
            return conditions;
        }

        public void setConditions(ArrayList<ExpressionPredicateCondition> conditions) {
            this.conditions = conditions;
        }

        public String getExpression() {
            return expression;
        }

        public void setExpression(String expression) {
            this.expression = expression;
        }

        public ArrayList<ExpressionPredicateCondition> conditions() {
            return conditions;
        }

        @Override
        public String toString() {
            StringBuilder conditionString = new StringBuilder();
            for(ExpressionPredicateCondition epc: conditions()) {
                conditionString.append(epc);
                conditionString.append(" ");
            }
            String output = String.format("ExpressionPredicate (\"%s\" -> %s)", getExpression(), conditionString);
            return output;
        }
    }

    public enum ExpressionPredicateOperator {
        NOT_NULL,
        EQUAL,
        LESS,
        GREATER,
        LESS_EQUAL,
        GREATER_EQUAL
    }

    public enum ExpressionPredicateCombineOperator {
        AND,
        OR,
    }

    /**
     * Complex Predicate Implementation.
     *
     * @author openecho
     */
    public static class ExpressionPredicateCondition {
        ExpressionPredicateCombineOperator combine;
        String attribute;
        ExpressionPredicateOperator operator;
        Value value;

        public ExpressionPredicateCondition() {
            combine = null;
            attribute = null;
            operator = null;
            value = null;
        }

        public ExpressionPredicateCondition(String attribute, ExpressionPredicateOperator operator) {
            this.attribute = attribute;
            this.operator = operator;
            value = null;
        }

        public ExpressionPredicateCondition(String attribute, ExpressionPredicateOperator operator, Value value) {
            this.attribute = attribute;
            this.operator = operator;
            this.value = value;
        }
        
        public ExpressionPredicateCondition(ExpressionPredicateCombineOperator combine, String attribute, ExpressionPredicateOperator operator, Value value) {
            this.combine = combine;
            this.attribute = attribute;
            this.operator = operator;
            this.value = value;
        }
       
        public ExpressionPredicateCombineOperator getCombine() {
            return combine;
        }

        public void setCombine(ExpressionPredicateCombineOperator combine) {
            this.combine = combine;
        }

        public String getAttribute() {
            return attribute;
        }

        public void setAttribute(String attribute) {
            this.attribute = attribute;
        }

        public ExpressionPredicateOperator getOperator() {
            return operator;
        }

        public void setOperator(ExpressionPredicateOperator operator) {
            this.operator = operator;
        }

        public Value getValue() {
            return value;
        }

        public void setValue(Value value) {
            this.value = value;
        }

        @Override
        public String toString() {
            return String.format("ExpressionPredicateCondition (%s%s %s %s)", ((combine != null) ? String.format("%s ", combine) : ""), attribute, operator, value.toJSON());
        }
    }
}
