/**
 * Copyright (C) 2010-2011 J.W.Marsden
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/
package jsonij.json;

import jsonij.json.JPath.ExpressionPredicateCombineOperator;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import jsonij.json.JPath.Component;
import jsonij.parser.ParserException;
import jsonij.parser.ReaderParser;

import static jsonij.json.Constants.MINUS;
import static jsonij.json.Constants.LEFT_SQUARE_BRACKET;
import static jsonij.json.Constants.RIGHT_SQUARE_BRACKET;
import static jsonij.json.Constants.LAST_CHAR;
import static jsonij.json.Constants.LAST_PREDICATE;
import static jsonij.json.Constants.SOLIDUS_CHAR;
import static jsonij.json.Constants.REVERSE_SOLIDUS;
import static jsonij.json.Constants.LEFT_PARENTHESIS;
import static jsonij.json.Constants.RIGHT_PARENTHESIS;
import static jsonij.json.Constants.EXPRESSION_CHAR;
import static jsonij.json.Constants.CURRENT_ELEMENT_CHAR;
import static jsonij.json.Constants.PERIOD_CHAR;
import static jsonij.json.Constants.LESS;
import static jsonij.json.Constants.GREATER;
import static jsonij.json.Constants.EQUAL;
import static jsonij.json.Constants.AND;
import static jsonij.json.Constants.OR;

/**
 *
 * @author openecho
 */
public class JPathParser {

    public JPath<Component> parse(String jPath) throws IOException, ParserException {
        Reader jPathStringReader = new StringReader(jPath);
        return parse(jPathStringReader);
    }

    public JPath<Component> parse(Reader jPathReader) throws IOException, ParserException {
        JPathReader target = new JPathReader(jPathReader);
        JPath<Component> path = new JPath<Component>();
        while (target.peek() != -1) {
            if (target.peek() == SOLIDUS_CHAR) {
                target.read();
            } else if (target.peek() == LEFT_SQUARE_BRACKET) {
                JPath.Predicate predicate = readPredicate(target);
                if (predicate != null) {
                    // TODO: test for complex duplication
                    path.add(predicate);
                }
            } else {
                JPath.Key key = readKey(target);
                if (key != null) {
                    path.add(key);
                }
            }
        }
        return path;
    }

    public JPath.Key readKey(JPathReader target) throws IOException, ParserException {
        StringBuilder keyBuilder = new StringBuilder();
        while (true) {
            if (target.peek() == -1) {
                break;
            } else if (target.peek() == SOLIDUS_CHAR) {
                break;
            } else if (target.peek() == LEFT_SQUARE_BRACKET) {
                break;
            } else if (target.peek() == REVERSE_SOLIDUS) {
                target.read();
                keyBuilder.append((char) target.read());
            } else {
                keyBuilder.append((char) target.read());
            }
        }
        return new JPath.Key(keyBuilder.toString());
    }

    public JPath.Predicate readPredicate(JPathReader target) throws IOException, ParserException {
        if (target.peek() == -1) {
            return null;
        }
        if (target.peek() == LEFT_SQUARE_BRACKET) {
            target.read();
        } else {
            throw new JPathParserException("invalidStringExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) '[', (char) target.peek());
        }

        StringBuilder predicateComponentBuilder = null;

        if (ConstantUtility.isDigit(target.peek())) {
            /**
             * Most likely case, first character is a digit and this is a SimpleIndex.
             **/
            predicateComponentBuilder = new StringBuilder();
            predicateComponentBuilder.append((char) target.read());
            while (ConstantUtility.isDigit(target.peek())) {
                predicateComponentBuilder.append((char) target.read());
            }
            if (target.peek() == RIGHT_SQUARE_BRACKET) {
                target.read();
                return new JPath.SimpleIndex(Integer.parseInt(predicateComponentBuilder.toString()));
            } else {
                throw new JPathParserException("invalidStringExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) ']', (char) target.peek());
            }
        } else if (target.peek() == LAST_CHAR) {
            /**
             * Check for $
             **/
            target.read();
            if (target.peek() == RIGHT_SQUARE_BRACKET) {
                target.read();
                return JPath.LastIndex.LAST;
            } else if (target.peek() == MINUS) {
                target.read();
                if (ConstantUtility.isDigit(target.peek())) {
                    predicateComponentBuilder = new StringBuilder();
                    predicateComponentBuilder.append((char) target.read());
                    while (ConstantUtility.isDigit(target.peek())) {
                        predicateComponentBuilder.append((char) target.read());
                    }
                    if (target.peek() == RIGHT_SQUARE_BRACKET) {
                        target.read();
                        return new JPath.LastIndex(Integer.parseInt(predicateComponentBuilder.toString()));
                    } else {
                        throw new JPathParserException("invalidStringExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) ']', (char) target.peek());
                    }
                } else {
                    throw new UnsupportedOperationException("Not a digit in LastIndex.");
                }
            } else {
                throw new JPathParserException("invalidValue", target.getLineNumber(), target.getPositionNumber(), (char) ']', (char) target.peek());
            }
        } else if (target.peek() == LAST_PREDICATE.charAt(0)) {
            /**
             * Check for last() predicate string
             **/
            for (int i = 0; i < LAST_PREDICATE.length(); i++) {
                if (target.peek() == LAST_PREDICATE.charAt(i)) {
                    target.read();
                } else {
                    throw new JPathParserException("invalidValue", target.getLineNumber(), target.getPositionNumber(), LAST_PREDICATE, (char) target.peek());
                }
            }
            if (target.peek() == RIGHT_SQUARE_BRACKET) {
                target.read();
                return JPath.LastIndex.LAST;
            } else if (target.peek() == MINUS) {
                target.read();
                if (ConstantUtility.isDigit(target.peek())) {
                    predicateComponentBuilder = new StringBuilder();
                    predicateComponentBuilder.append((char) target.read());
                    while (ConstantUtility.isDigit(target.peek())) {
                        predicateComponentBuilder.append((char) target.read());
                    }
                    if (target.peek() == RIGHT_SQUARE_BRACKET) {
                        target.read();
                        return new JPath.LastIndex(Integer.parseInt(predicateComponentBuilder.toString()));
                    } else {
                        throw new JPathParserException("invalidStringExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) ']', (char) target.peek());
                    }
                } else {
                    throw new UnsupportedOperationException("Not a digit in LastIndex.");
                }
            } else {
                throw new JPathParserException("invalidPathExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) ']', (char) target.peek());
            }
        } else if(target.peek() == EXPRESSION_CHAR) {
            JPath.ExpressionPredicate expressionPredicate = new JPath.ExpressionPredicate();

            target.read();
            if (target.peek() == LEFT_PARENTHESIS) {
                target.read();

                StringBuilder expressionStringBuilder = new StringBuilder();
                char c;
                while (target.peek() != RIGHT_PARENTHESIS) {
                    if(target.peek() == -1) {
                        throw new JPathParserException("invalidPathFoundEndExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) ')', (char) target.peek());
                    }

                    target.skipWhitepace(expressionStringBuilder);
                    
                    ExpressionPredicateCombineOperator op = null;
                    
                    if(target.peek() == AND) {
                        while(target.peek() == AND) {
                            expressionStringBuilder.append((char) target.read());
                        }
                        op = ExpressionPredicateCombineOperator.AND;
                    } else if(target.peek() == OR) {
                        while(target.peek() == OR) {
                            expressionStringBuilder.append((char) target.read());
                        }
                        op = ExpressionPredicateCombineOperator.OR;
                    }
                    target.skipWhitepace(expressionStringBuilder);
                    if (target.peek() == CURRENT_ELEMENT_CHAR) {
                        expressionStringBuilder.append((char) target.read());
                        if(target.peek() == PERIOD_CHAR) {
                            expressionStringBuilder.append((char) target.read());

                            target.skipWhitepace(expressionStringBuilder);
                    
                            String attribute = null;
                            String value = null;
                            
                            predicateComponentBuilder = new StringBuilder();

                            while(target.peek() != RIGHT_PARENTHESIS && target.peek() != LESS && target.peek() != GREATER && target.peek() != EQUAL) {
                                c = (char) target.read();
                                expressionStringBuilder.append(c);
                                predicateComponentBuilder.append(c);
                            }
                            attribute = predicateComponentBuilder.toString().trim();

                            JPath.ExpressionPredicateOperator expressionPredicateOperator = null;
                            if(target.peek() == RIGHT_PARENTHESIS) {
                                expressionStringBuilder.append((char) target.read());
                                expressionPredicateOperator = JPath.ExpressionPredicateOperator.NOT_NULL;
                            } else if(target.peek() == EQUAL) {
                                expressionStringBuilder.append((char) target.read());
                                expressionPredicateOperator = JPath.ExpressionPredicateOperator.EQUAL;
                            } else if(target.peek() == LESS) {
                                expressionStringBuilder.append((char) target.read());
                                if(target.peek() == EQUAL) {
                                    expressionStringBuilder.append((char) target.read());
                                    expressionPredicateOperator = JPath.ExpressionPredicateOperator.LESS_EQUAL;
                                } else {
                                    expressionPredicateOperator = JPath.ExpressionPredicateOperator.LESS;
                                }
                            } else if(target.peek() == GREATER) {
                                expressionStringBuilder.append((char) target.read());
                                if(target.peek() == EQUAL) {
                                    expressionStringBuilder.append((char) target.read());
                                    expressionPredicateOperator = JPath.ExpressionPredicateOperator.GREATER_EQUAL;
                                } else {
                                    expressionPredicateOperator = JPath.ExpressionPredicateOperator.GREATER;
                                }
                            }
                            target.skipWhitepace(expressionStringBuilder);
                            predicateComponentBuilder = new StringBuilder();
                            while(target.peek() != RIGHT_PARENTHESIS && target.peek() != AND && target.peek() != OR) {
                                if(target.peek() == -1) {
                                    throw new JPathParserException("invalidPathFoundEndExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) ')', (char) target.peek());
                                }
                                c = (char) target.read();
                                expressionStringBuilder.append(c);
                                predicateComponentBuilder.append(c);
                            }
                            value = predicateComponentBuilder.toString().trim();
                            JSONParser jsonParser = new JSONParser();
                            Value jsonValue = jsonParser.parseValue(value);
                            target.skipWhitepace(expressionStringBuilder);
                            JPath.ExpressionPredicateCondition predicateCondition = new JPath.ExpressionPredicateCondition(attribute, expressionPredicateOperator, jsonValue);
                            if(op != null) {
                                predicateCondition.setCombine(op);
                                op = null;
                            }
                            expressionPredicate.conditions().add(predicateCondition);
                        }
                    } else {
                        throw new JPathParserException("expression1", target.getPositionNumber());
                    }
                }
                // Read Left Parenthesis
                target.read();

                expressionPredicate.setExpression(expressionStringBuilder.toString());

                if(target.peek() == RIGHT_SQUARE_BRACKET) {
                    expressionStringBuilder.append((char) target.read());
                }

                return expressionPredicate;
            } else {
                throw new JPathParserException("invalidStringExpecting1", target.getLineNumber(), target.getPositionNumber(), (char) '(', (char) target.peek());
            }
        }
        throw new UnsupportedOperationException("Invalid Predicate.");
    }

    public static class JPathReader extends ReaderParser {

        public JPathReader() {
            this(null);
        }

        public JPathReader(Reader reader) {
            super(reader);
        }

        @Override
        protected int readNext() throws IOException, ParserException {
            return getReader().read();
        }

        public void skipWhitepace() throws IOException, ParserException {
            while(ConstantUtility.isWhiteSpace(peek())) {
                read();
            }
        }

        public void skipWhitepace(StringBuilder appender) throws IOException, ParserException {
            while(ConstantUtility.isWhiteSpace(peek())) {
                appender.append((char) read());
            }
        }
    }
}
