/**
 * Copyright (C) 2010-2011 J.W.Marsden
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/
package jsonij.json;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import jsonij.json.annotation.JSONEncoder;
import java.io.IOException;
import java.io.Reader;
import jsonij.json.Inspector.InspectorProperty;
import jsonij.parser.ParserException;

/**
 *
 * @author openecho
 */
public class JSONMarshaler {

    static JSONCodecStore codecStore;

    public enum TYPE {
        BOOLEAN,
        NUMERIC,
        STRING,
        ARRAY,
        LIST,
        OBJECT,
        MAP,
        NA
    }

    /**
     * Java Object Marshal Access Points.
     */
    /**
     *
     * @param o
     * @return
     */
    public static JSON marshalObject(Object o) {
        Value object = null;
        if (o instanceof List) {
            object = marshalListType(o);
        } else if (o instanceof Map) {
            object = marshalMapType(o);
        } else {
            object = JSONMarshaler.marshalObjectType(o);
        }
        if (object == null) {
            object = new JSON.Object<JSON.String, Value>();
        }
        return new JSON(object);
    }

    public static JSON marshalObject(boolean[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Boolean> jsonArray = new JSON.Array<JSON.Boolean>();
        for (int i = 0; i < size; i++) {
            if (a[i]) {
                jsonArray.add(JSON.TRUE);
            } else {
                jsonArray.add(JSON.FALSE);
            }
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Boolean[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Boolean> jsonArray = new JSON.Array<JSON.Boolean>();
        for (int i = 0; i < size; i++) {
            if (a[i]) {
                jsonArray.add(JSON.TRUE);
            } else {
                jsonArray.add(JSON.FALSE);
            }
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(int[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Integer[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(char[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.String> jsonArray = new JSON.Array<JSON.String>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.String("" + a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Character[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.String> jsonArray = new JSON.Array<JSON.String>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.String("" + a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(double[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Double[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(float[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Float[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(short[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Short[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(long[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Long[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Numeric> jsonArray = new JSON.Array<JSON.Numeric>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.Numeric(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(String[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.String> jsonArray = new JSON.Array<JSON.String>();
        for (int i = 0; i < size; i++) {
            jsonArray.add(new JSON.String(a[i]));
        }
        return new JSON(jsonArray);
    }

    public static JSON marshalObject(Object[] a) {
        int size = 0;
        if ((size = Array.getLength(a)) == 0) {
            return new JSON(new JSON.Array<JSON.Numeric>());
        }
        JSON.Array<JSON.Object> jsonArray = new JSON.Array<JSON.Object>();
        for (int i = 0; i < size; i++) {
            jsonArray.add((JSON.Object) marshalObjectType(a[i]));
        }
        return new JSON(jsonArray);
    }

    /**
     * JSON Object Marshal Access Points.
     */
    /**
     * Helper Methods
     */
    /**
     * Marshals any Java Object into the best Value possible. This method inspects
     * the Object's type and then passes it to one of the other helper methods
     * to marshal.
     * @param o The object to inspect and marshal.
     * @return The best Value type to represent the object.
     */
    public static Value marshalAnyObject(Object o) {
        Value value = null;
        if (o == null) {
            return JSON.NULL;
        }
        Class<?> marshaledObjectClass = o.getClass();
        TYPE type;
        if ((type = isBooleanJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalBooleanType(o);
        } else if ((type = isNumericJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalNumericType(o);
        } else if ((type = isStringJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalStringType(o);
        } else if ((type = isArrayJSONType(marshaledObjectClass)) != TYPE.NA) {
            if (type == TYPE.ARRAY) {
                value = marshalArrayType(o);
            } else if (type == TYPE.LIST) {
                value = marshalListType(o);
            }
        } else if ((type = isObjectJSONType(marshaledObjectClass)) != TYPE.NA) {
            if (type == TYPE.MAP) {
                value = marshalMapType(o);
            } else if (type == TYPE.OBJECT) {
                value = marshalObjectType(o);
            }
        } else {
            value = new JSON.String(o.getClass().getCanonicalName() + "@" + Integer.toHexString(o.hashCode()));
        }
        return value;
    }

    public static TYPE isBooleanJSONType(Class<?> c) {
        return (c == Boolean.class || c == boolean.class) ? TYPE.BOOLEAN : TYPE.NA;
    }

    public static TYPE isNumericJSONType(Class<?> c) {
        if (c == int.class || c == Integer.class) {
            return TYPE.NUMERIC;
        }
        if (c == double.class || c == Double.class) {
            return TYPE.NUMERIC;
        }
        if (c == float.class || c == Float.class) {
            return TYPE.NUMERIC;
        }
        if (c == long.class || c == Long.class) {
            return TYPE.NUMERIC;
        }
        if (c == short.class || c == Short.class) {
            return TYPE.NUMERIC;
        }
        return TYPE.NA;
    }

    public static TYPE isStringJSONType(Class<?> c) {
        if (c == String.class) {
            return TYPE.STRING;
        } else if (c == char.class || c == Character.class) {
            return TYPE.STRING;
        }
        return TYPE.NA;
    }

    public static TYPE isArrayJSONType(Class<?> c) {
        if (c.isArray()) {
            return TYPE.ARRAY;
        }
        Class<?>[] interfaces = c.getInterfaces();
        for (int i = 0; i < Array.getLength(interfaces); i++) {
            if (interfaces[i] == List.class) {
                return TYPE.LIST;
            }
        }
        return TYPE.NA;
    }

    public static TYPE isObjectJSONType(Class<?> c) {
        Inspector inspect = new Inspector(c);
        inspect.inspect();
        InspectorProperty[] inspectProperties = inspect.getProperties();
        if (inspectProperties != null) {
            for (int i = 0; i < Array.getLength(inspectProperties); i++) {
                if (inspectProperties[i].isValid()) {
                    return TYPE.OBJECT;
                }
            }
        }
        return TYPE.NA;
    }

    public static Value marshalBooleanType(Object o) {
        Value value = null;
        boolean booleanValue = (Boolean) o;
        if (booleanValue) {
            value = JSON.TRUE;
        } else {
            value = JSON.FALSE;
        }
        return value;
    }

    public static Value marshalNumericType(Object o) {
        Value value = null;
        Number marshaledNumber = null;
        marshaledNumber = (Number) o;
        if (marshaledNumber != null) {
            value = new JSON.Numeric(marshaledNumber);
        } else {
            value = JSON.NULL;
        }
        return value;
    }

    public static Value marshalStringType(Object o) {
        Value value = null;
        String marshaledString = null;
        marshaledString = (String) o;
        if (marshaledString != null) {
            value = new JSON.String(marshaledString);
        } else {
            value = JSON.NULL;
        }
        return value;
    }

    public static Value marshalArrayType(Object o) {
        Value value = null;
        Object marshaledArray = o;
        if (marshaledArray != null) {
            JSON.Array<Value> jsonArray = new JSON.Array<Value>();
            int size = Array.getLength(marshaledArray);
            Class<?> type = marshaledArray.getClass();
            Class<?> componentType = type.getComponentType();
            if (componentType == int.class || componentType == Integer.class) {
                for (int i = 0; i < size; i++) {
                    jsonArray.add(new JSON.Numeric(Array.getInt(marshaledArray, i)));
                }
            } else if (componentType == double.class || componentType == Double.class) {
                for (int i = 0; i < size; i++) {
                    jsonArray.add(new JSON.Numeric(Array.getDouble(marshaledArray, i)));
                }
            } else if (componentType == float.class || componentType == Float.class) {
                for (int i = 0; i < size; i++) {
                    jsonArray.add(new JSON.Numeric(Array.getFloat(marshaledArray, i)));
                }
            } else if (componentType == short.class || componentType == Short.class) {
                for (int i = 0; i < size; i++) {
                    jsonArray.add(new JSON.Numeric(Array.getShort(marshaledArray, i)));
                }
            } else if (componentType == long.class || componentType == Long.class) {
                for (int i = 0; i < size; i++) {
                    jsonArray.add(new JSON.Numeric(Array.getLong(marshaledArray, i)));
                }
            } else {
                for (int i = 0; i < size; i++) {
                    Object arrayValue = Array.get(marshaledArray, i);
                    jsonArray.add(marshalAnyObject(arrayValue));
                }
            }
            value = jsonArray;
        } else {
            value = JSON.NULL;
        }
        return value;
    }

    public static Value marshalListType(Object o) {
        Value value = null;
        List<?> marshaledList = (List<?>) o;
        if (marshaledList != null) {
            JSON.Array<Value> jsonArray = new JSON.Array<Value>();
            Iterator<?> marshaledListIterator = marshaledList.listIterator();
            Object listItem = null;
            while (marshaledListIterator.hasNext()) {
                listItem = marshaledListIterator.next();
                if (listItem == null) {
                    jsonArray.add(JSON.NULL);
                    continue;
                }
                jsonArray.add(marshalAnyObject(listItem));
            }
            value = jsonArray;
        } else {
            value = JSON.NULL;
        }
        return value;
    }

    /**
     * Marshal an Object by considering its methods and attributes. This converts
     * the passed object into the best JSON Object representation possible by
     * reflecting the getters, setters and public attributes.
     *
     * @param o Object to marshal.
     * @return Value for the JSON.Object
     */
    public static Value marshalObjectType(Object o) {
        Class<?> objectClass = o.getClass();
        // Check for JSONEncoder Annotation
        Method[] methods = objectClass.getDeclaredMethods();
        for (Method method : methods) {
            if (method.getAnnotation(JSONEncoder.class) != null) {
            }
        }

        // Check for JSONCodec
        if (codecStore != null && codecStore.hasCodec(objectClass)) {
            JSONCodec codec = codecStore.getCodec(objectClass);
            Value value = codec.encode(o);
            return value;
        }

        // Inspect Marshal the JSON Object
        JSON.Object<JSON.String, Value> jsonObject = new JSON.Object<JSON.String, Value>();
        Inspector inspector = new Inspector(o);
        inspector.inspect();
        Inspector.InspectorProperty[] properties = inspector.getProperties();
        String name = null;
        Value value = null;
        for (Inspector.InspectorProperty property : properties) {
            if (property.isIgnore() || !property.isValid()) {
                continue;
            }
            name = property.getName();
            if (property.getAccessType() == Inspector.InspectorProperty.TYPE.METHOD) {
                String accessorName = property.getAccessName();
                try {
                    Method method = objectClass.getMethod(accessorName);
                    value = JSONMarshaler.marshalObjectMethodValue(method, o);
                } catch (NoSuchMethodException ex) {
                    value = new JSON.String(ex.toString());
                } catch (SecurityException ex) {
                    value = new JSON.String(ex.toString());
                }
            } else if (property.getAccessType() == Inspector.InspectorProperty.TYPE.FIELD) {
                String accessorName = property.getAccessName();
                try {
                    Field field = objectClass.getField(accessorName);
                    value = JSONMarshaler.marshalObjectFieldValue(field, o);
                } catch (NoSuchFieldException ex) {
                    value = new JSON.String(ex.toString());
                } catch (SecurityException ex) {
                    value = new JSON.String(ex.toString());
                }

            } else {
                value = JSON.NULL;
            }
            jsonObject.put(new JSON.String(name), value);
        }
        return jsonObject;
    }

    public static Value marshalMapType(Object o) {
        Value value = null;
        Map<?, ?> marshaledMap = (Map<?, ?>) o;
        if (marshaledMap != null) {
            JSON.Object<JSON.String, Value> jsonObject = new JSON.Object<JSON.String, Value>();
            Iterator<?> keySetIterator = marshaledMap.keySet().iterator();
            while (keySetIterator.hasNext()) {
                Object keyObject = keySetIterator.next();
                Object valueObject = marshaledMap.get(keyObject);
                jsonObject.put(new JSON.String(keyObject.toString()), marshalAnyObject(valueObject));
            }
            value = jsonObject;
        } else {
            value = JSON.NULL;
        }
        return value;
    }

    public static Value marshalObjectMethodValue(Method method, Object o) {
        Value value = null;
        Class<?> marshaledObjectClass = method.getReturnType();
        TYPE type;

        Object marshaledObject = null;
        try {
            marshaledObject = method.invoke(o);
        } catch (IllegalAccessException ex) {
            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        } catch (IllegalArgumentException ex) {
            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        } catch (InvocationTargetException ex) {
            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        }


        if ((type = JSONMarshaler.isBooleanJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalBooleanType(marshaledObject);
        } else if ((type = JSONMarshaler.isNumericJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalNumericType(marshaledObject);
        } else if ((type = JSONMarshaler.isStringJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalStringType(marshaledObject);
        } else if ((type = JSONMarshaler.isArrayJSONType(marshaledObjectClass)) != TYPE.NA) {
            if (type == TYPE.ARRAY) {
                value = marshalArrayType(marshaledObject);
            } else if (type == TYPE.LIST) {
                value = marshalListType(marshaledObject);
            }
        } else if ((type = JSONMarshaler.isObjectJSONType(marshaledObjectClass)) != TYPE.NA) {
            if (type == TYPE.MAP) {
                value = new JSON.String("Map!");
            } else if (type == TYPE.OBJECT) {

                value = JSONMarshaler.marshalObjectType(marshaledObject);
            }
        } else {
            if (marshaledObject != null) {
                value = new JSON.String(marshaledObject.getClass().getCanonicalName() + "@" + Integer.toHexString(marshaledObject.hashCode()));
            } else {
                value = new JSON.String(String.format("UKNOWN Value for field %s", method.getName()));
            }
        }
        return value;
    }

    public static Value marshalObjectFieldValue(Field field, Object o) {
        Value value = null;
        Class<?> marshaledObjectClass = field.getType();
        TYPE type;

        Object marshaledObject = null;
        try {
            marshaledObject = field.get(o);
        } catch (IllegalArgumentException ex) {
            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        } catch (IllegalAccessException ex) {
            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        }

        if ((type = JSONMarshaler.isBooleanJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalBooleanType(marshaledObject);
        } else if ((type = JSONMarshaler.isNumericJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalNumericType(marshaledObject);
        } else if ((type = JSONMarshaler.isStringJSONType(marshaledObjectClass)) != TYPE.NA) {
            value = marshalStringType(marshaledObject);
        } else if ((type = JSONMarshaler.isArrayJSONType(marshaledObjectClass)) != TYPE.NA) {
            if (type == TYPE.ARRAY) {
                value = marshalArrayType(marshaledObject);
            } else if (type == TYPE.LIST) {
                value = marshalListType(marshaledObject);
            }
        } else if ((type = JSONMarshaler.isObjectJSONType(marshaledObjectClass)) != TYPE.NA) {
            if (type == TYPE.MAP) {
                value = marshalMapType(marshaledObject);
            } else if (type == TYPE.OBJECT) {

                value = JSONMarshaler.marshalObjectType(marshaledObject);
            }
        } else {
            if (marshaledObject != null) {
                value = new JSON.String(marshaledObject.getClass().getCanonicalName() + "@" + Integer.toHexString(marshaledObject.hashCode()));
            } else {
                value = new JSON.String(String.format("UKNOWN Value for field %s", field.getName()));
            }
        }
        return value;
    }

    public static Object marshalJSON(String jsonString, Class<?> clazz) throws ParserException, IOException {
        JSON json = JSON.parse(jsonString);
        return marshalJSON(json, clazz);
    }

    public static Object marshalJSON(Reader jsonReader, Class<?> clazz) throws ParserException, IOException {
        JSON json = JSON.parse(jsonReader);
        return marshalJSON(json, clazz);
    }

    public static Object marshalJSON(JSON json, Class<?> clazz) {
        return marshalJSON(json.getRoot(), clazz);
    }

    public static Object marshalJSON(Value value, Class<?> clazz) {
        Object objectInstance = null;
        Value.TYPE valueType = value.getValueType();
        if (valueType == Value.TYPE.ARRAY) {
            // TODO: Check Array Types.

            return 0;
        } else if (valueType == Value.TYPE.OBJECT) {
            try {
                objectInstance = clazz.newInstance();
            } catch (InstantiationException ex) {
                Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IllegalAccessException ex) {
                Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
            }
            if(objectInstance == null) {
                // todo: exception
            }
            marshalPropertyValues(clazz, objectInstance, value);

            return objectInstance;
        } else {
            throw new UnsupportedOperationException("Unknown Type. Have you made a new one?");
        }
    }

    public static void marshalPropertyValues(Class<?> clazz, Object o, Value value) {
        Inspector inspector = new Inspector(clazz);
            inspector.inspect();
            InspectorProperty[] properties = inspector.getProperties();
            for (InspectorProperty property : properties) {
                System.out.println("Prop: " + property);
                if(property.isValid() && value.has(property.getName())) {
                    System.out.println("Found");
                    Value sourceValue = value.get(property.getName());

                    if(property.getAccessType() == Inspector.InspectorProperty.TYPE.FIELD) {

                    } else if(property.getAccessType() == Inspector.InspectorProperty.TYPE.METHOD) {

                        Class<?> parmType = null;
                        Object parmValue = null;
                        if(sourceValue.type() == Value.TYPE.NULL) {
                            // Nothing to do here.
                            continue;
                        } else if(sourceValue.type() == Value.TYPE.TRUE) {
                            parmType = Boolean.class;
                            parmValue = true;
                        } else if(sourceValue.type() == Value.TYPE.FALSE) {
                            parmType = Boolean.class;
                            parmValue = false;
                        } else if(sourceValue.type() == Value.TYPE.NUMERIC) {
                            parmType = Number.class;
                            parmValue = sourceValue.getNumber();
                        } else if(sourceValue.type() == Value.TYPE.STRING) {
                            parmType = String.class;
                            parmValue = sourceValue.getString();
                        } else if(sourceValue.type() == Value.TYPE.ARRAY) {
                            parmType = Array.class;

                        } else if(sourceValue.type() == Value.TYPE.OBJECT) {
                            parmType = Object.class;
                            
                        }

                        try {
                            Method method = clazz.getMethod(property.getMutateName(), parmType);
                            method.invoke(o, parmValue);
                        } catch (IllegalAccessException ex) {
                            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
                        } catch (IllegalArgumentException ex) {
                            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
                        } catch (InvocationTargetException ex) {
                            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
                        } catch (NoSuchMethodException ex) {
                            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
                        } catch (SecurityException ex) {
                            Logger.getLogger(JSONMarshaler.class.getName()).log(Level.SEVERE, null, ex);
                        }
                    } else {
                        // todo: throw exception.
                    }
                }
            }
    }

    public static void setMethodValue(Method method, Value value) {

    }

    public static boolean hasCodec(Class codecType) {
        if (codecStore == null) {
            return false;
        } else {
            return codecStore.hasCodec(codecType);
        }
    }

    public static void registerCodec(Class<?> codecClass) {
        if (codecStore == null) {
            codecStore = new JSONCodecStore();
        }
        codecStore.registerCodec(codecClass);
    }

    public static JSONCodec<?> getCodec(Class codecType) {
        if (codecStore == null) {
            return null;
        } else {
            return codecStore.getCodec(codecType);
        }
    }
}
