/**
 * Copyright (C) 2010-2011 J.W.Marsden
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/
package jsonij.json;

import java.io.IOException;
import java.io.Reader;
import java.util.logging.Level;
import java.util.logging.Logger;
import jsonij.parser.ReaderParser;

public class JSONReader extends ReaderParser {

    JSONStringReader stringReader;

    static final Logger logger = Logger.getLogger(JSONReader.class.getName());

    public JSONReader() {
        this(null);
    }

    public JSONReader(Reader reader) {
        super(reader);
        stringReader = null;
    }

    /**
     * Reads from the reader.
     *
     * @param targetReader The reader to be read from.
     * @return The read byte if found otherwise -1 if the end of the stream is reached.
     * @throws IOException Java IO Exception.
     */
    protected int readNext() throws IOException {
        if (logger.isLoggable(Level.FINER)) {
            logger.entering(getClass().getName(), "readIgnoreWhiteSpace()");
        }
        if (reader == null) {
            throw new NullPointerException("Null Reader Exception.");
        }
        int r = -1;
        while (reader.ready() && (r = reader.read()) != -1) {
            getPosition().tickReadCount();
            getPosition().movePosition();
            if (ConstantUtility.isNewLine(r)) {
                handleNewLine();
            } else if (!ConstantUtility.isWhiteSpace(r)) {
                break;
            }
        }
        if (logger.isLoggable(Level.FINER)) {
            logger.exiting(getClass().getName(), "readIgnoreWhiteSpace()", r);
        }
        return r;
    }

    public JSONStringReader getStringReader() {
        if (stringReader == null) {
            stringReader = new JSONStringReader();
        }
        return stringReader;
    }

    protected class JSONStringReader {

        protected boolean active;

        public JSONStringReader() {
            active = true;
        }

        /**
         * @return the active
         */
        public boolean isActive() {
            return active;
        }

        /**
         * @param active the active to set
         */
        protected void setActive(boolean active) {
            this.active = active;
        }

        public int peek() throws IOException {
            if (stringReader == null || !active) {
                // TODO: Dead Reader!
            }
            if (!hasPeeked) {
                peekValue = readNext();
                hasPeeked = true;
            }
            return peekValue;
        }

        public int read() throws IOException {
            if (stringReader == null || !active) {
                // TODO: Dead Reader!
            }
            if (hasPeeked) {
                hasPeeked = false;
                return peekValue;
            }
            return readNext();
        }

        public void close() throws IOException {
            stringReader = null;
            active = false;
        }

        protected int readNext() throws IOException {
            String method = "readIgnoreWhiteSpace()";
            if (logger.isLoggable(Level.FINER)) {
                logger.entering(getClass().getName(), method);
            }
            if (reader == null) {
                throw new NullPointerException("Null Reader Exception.");
            }
            int r = -1;
            while (reader.ready() && (r = reader.read()) != -1) {
                getPosition().tickReadCount();
                getPosition().movePosition();
                if (ConstantUtility.isNewLine(r)) {
                    handleNewLine();
                } else {
                    break;
                }
            }
            if (logger.isLoggable(Level.FINER)) {
                logger.exiting(getClass().getName(), method, r);
            }
            return r;
        }
    }
}
